module Grays42_GPath

=begin
	Entity_Path is similar to a PGL_Path, except it is equipped to be handed Sketchup geometry items instead of PGL items.
	It converts existing sketchup entities into PGL items and loads them into a PGL path.
=end
	class Entity_Path < PGL_Path
		
		def initialize(array = [])
			super(array)
			@parent_curves = []
		end
		
		def <<(entity)
			if entity.is_a?(Sketchup::Entity) then
				append_entity(entity)
			elsif entity.is_a?(Array) then
				entity.each do |e|
					append_entity(e)
				end
			elsif entity.is_a?(PGL_Item) then
				append_item(entity)
			else
				raise "Error: item of class #{entity.class} cannot be attached to an Entity_Path."
			end
		end
		
		#NOTE: APPENDING AN ENTITY TO AN ENTITY PATH ASSUMES THAT THE ENTITY CONNECTS!! Will throw crazy errors if it doesn't.
		def append_entity(entity,transformation = Geom::Transformation.new)
			case entity
				when Sketchup::Edge
					append_edge(entity, transformation) if entity.curve.nil?
					append_arccurve(entity, transformation) if entity.curve.class == Sketchup::ArcCurve
				else
					puts "Warning: cannot append entity of class #{entity.class}." 
			end
		end
		
		
		def append_edge(edge, transformation)
			start_pt = edge.start.position.transform(transformation)
			end_pt = edge.end.position.transform(transformation)
			pgl = PGL_SegmentDef.new(start_pt,end_pt)
			pgl = check_for_edge_reversal(pgl)
			append_item(pgl, nil, true)
		end
		
		#Takes in the nonsense xaxis/start_angle or xaxis/end_angle notation and returns a proper vector.
		def vector_from_axis(xaxis,angle,normal)
			# http://stackoverflow.com/questions/620745/c-rotating-a-vector-around-a-certain-point
			angle = angle * normal.normalize.z
			xaxis = de_float_error(xaxis)
			new_x = de_float_error(((xaxis.x)*Math::cos(angle) - (xaxis.y)*Math::sin(angle)))
			new_y = de_float_error(((xaxis.y)*Math::cos(angle) + (xaxis.x)*Math::sin(angle)))
			new_z = de_float_error(xaxis.z)
			return Geom::Vector3d.new(new_x,new_y,new_z)
		end
		
		def append_arccurve(edge, transformation)
			raise "This is not a curve" unless edge.curve
			return if @parent_curves.include?(edge.curve)
			@parent_curves << edge.curve
			
			c = edge.curve.center.transform(transformation)
			r = edge.curve.radius.to_f
			sep = edge.length
			n = de_float_error(edge.curve.normal.transform(transformation))
			
			xaxis = edge.curve.xaxis.transform(transformation)
			v1 = vector_from_axis(xaxis,edge.curve.start_angle,n)
			v2 = vector_from_axis(xaxis,edge.curve.end_angle,n)
			
			arc = PGL_ArcDef.new(c, r, v1,v2,n,sep)
			arc = check_for_edge_reversal(arc)
	
			append_item(arc, nil, true)
		end
		
		#When checking Sketchup loops around faces, the vertices for the edges are NOT necessarily in the appropriate order.
		#This routine checks the endpoints of the edges to flip the entity around if necessary.
		def check_for_edge_reversal(pgl_item)
			return pgl_item if array_length == 0
			pgl_item = special_case_first_join(pgl_item) if array_length == 1
			return pgl_item.reverse if end_pt.distance(pgl_item.end_pt) < end_pt.distance(pgl_item.start_pt)
			return pgl_item 
		end
		
		#In attaching the second piece to the first piece, we not only don't know whether the second piece needs to be flipped, we also don't
		#know whether the first piece needs to be flipped. This processes both.
		def special_case_first_join(pgl)
			
			flip_0 = false
			flip_1 = false
			distance = end_pt.distance(pgl.start_pt)
			if end_pt.distance(pgl.end_pt) < distance then
				flip_0 = false
				flip_1 = true
				distance = end_pt.distance(pgl.end_pt)
			end
			if start_pt.distance(pgl.start_pt) < distance then
				flip_0 = true
				flip_1 = false
				distance = start_pt.distance(pgl.start_pt)
			end
			if start_pt.distance(pgl.end_pt) < distance then
				flip_0 = true
				flip_1 = true
				distance = start_pt.distance(pgl.end_pt)
			end
			self.last.reverse! if flip_0
			return pgl.reverse if flip_1
			return pgl
		end
	end

end
