module Grays42_GPath

	class Environment_Definition
		
		#The size of the bit
		attr_accessor :tool_radius
		
		#The radius of all interior corners so that the bit runs more smoothly
		attr_accessor :interior_radius
		
		#An array of z values that the bit will work through
		attr_accessor :cut_z_array
		
		#True if the bit will descend slowly through the piece, incrementing down smoothly through all layers.
		#False for traditional "rigid" descent
		attr_accessor :gradual_descent
		
		#Not currently used, will be a bounding box
		attr_accessor :workspace
		
		#The units/min rate of feeding horizontally through material
		attr_accessor :feed_rate
		
		#The maximum feed rate at which the tool can move *down*into the material without accounting for horizontal movement.
		attr_accessor :vertical_feed_rate
		
		#The units/min rate of seeking through empty space
		attr_accessor :seek_rate
		
		#The z at which the tool can seek safely without clipping leftover material
		attr_accessor :safe_height
		
		#The time t in seconds for a dwell
		attr_accessor :dwell_time
		
		#The position the machine should start and end at (Point3d)
		attr_accessor :home_x
		attr_accessor :home_y
		attr_accessor :home_z
		
		#The "clearance height" of the material, i.e. the height at which the tooltip is JUST above the material.
		#This is NOT the safe height at which to seek around; this is only used to seek straight down to the bed prior
		#to beginning a cut.
		attr_accessor :material_clearance
		
		#Screws are a way to mark spots to anchor down the material prior to the regular cut. Make any number of groups
		#or component instances, each named "SCREW", and the screw code will drill down to this depth for each. Perform
		#this gcode job prior to the regular gcode job.
		attr_accessor :screw_depth
		
		
		def validate!
			numeric = Object::Numeric
			assert_class(:tool_radius,numeric)
			assert_class(:interior_radius,numeric)
			assert_class(:cut_z_array,Array)
			0.upto(@cut_z_array.length - 1) do |i|
				next if @cut_z_array[i].is_a? numeric
				error_class("cut_z_array[#{i}]",@cut_z_array[i],@cut_z_array[i].class,numeric)
			end
			error_class(:gradual_descent,@gradual_descent,@gradual_descent.class,"true or false") unless @gradual_descent == true or @gradual_descent == false
			#NOTE: not checking workspace yet!
			assert_class(:feed_rate,numeric)
			assert_class(:vertical_feed_rate,numeric)
			assert_class(:seek_rate,numeric)
			assert_class(:safe_height,numeric)
			assert_class(:dwell_time,numeric)
			assert_class(:home_x,numeric)
			assert_class(:home_y,numeric)
			assert_class(:home_z,numeric)
			assert_class(:material_clearance,numeric)
			assert_class(:screw_depth,numeric)
		end
			
		def error_class(name,value,actual_class,expected_class)
			raise "Error: variable #{name} (value: #{value}) of class #{actual_class}, expected #{expected_class}"
		end
		
		def assert_class(variable_name,expected_class)
			actual_variable = self.method(variable_name).call
			return if actual_variable.is_a? expected_class
			error_class(variable_name,actual_variable,actual_variable.class,expected_class)
		end
		
		def valid?
			return true
		end
		
		def pretty_string
			str =   "Tool Radius #{tool_radius} (Interior Radius #{interior_radius})\n"
			str << "Cut at Z = #{cut_z_array.inspect}, gradually? #{gradual_descent}\n"
			str << "Seek #{seek_rate}/min, Feed #{feed_rate}/min, Max Vertical Feed #{vertical_feed_rate}/min\n"
			str << "Safe Seek Height #{safe_height}, Flush Material Clearance #{material_clearance}, Screw Depth #{screw_depth}\n"
			str << "Home Position #{home_x}, #{home_y}, #{home_z}, Dwell Length #{dwell_time} sec."
			return str
		end
		
		def instance_variables_hash
			#Thanks, Sketchup, for not being able to save a standard hash to a dictionary variable, I love making really crappy workarounds like this.
			hash_sorta = [ [] , [] ]
			keys = hash_sorta[0]
			values = hash_sorta[1]
			instance_variables.each do |name|
				keys << name
				values << instance_variable_get(name)
			end
			return hash_sorta
		end
		
		def unpack_variables_hash(hash_sorta)
			raise "Error: can only restore from a hash-like array created by Environment_Definition.instance_variables_hash" unless hash_sorta.length == 2 and hash_sorta[0].class == Array and hash_sorta[1].class == Array and hash_sorta[0].length == hash_sorta[1].length
			begin
				0.upto(hash_sorta[0].length - 1) do |i|
					instance_variable_set(hash_sorta[0][i],hash_sorta[1][i])
				end
			rescue
				raise "Error: could not unpack the variables hash.\n\n#{hash_sorta.to_s}"
			end
			return true
		end
		
		def self.valid_stored_env?(stored_env)
			return false unless stored_env.class == Array
			return false unless stored_env.length == 2
			keys = stored_env[0]
			values = stored_env[1]
			return false unless keys.class == Array and values.class == Array
			return false unless keys.length == values.length
			keys.each do |k|
				return false unless k.class == String
				method_name = k.reverse.chop.reverse
				return false unless instance_methods.include?(method_name)
			end
			return true
		end
		
	end
end