module Grays42_GPath

	class Gcode_Generator
		
		def initialize(environment)
			@env = environment
			@unit_mode = :inches
			@gcode_out = ""
			@header_is_set = false
			
			@total_feed_distance = 0
			@total_seek_distance = 0
			@current_feed_rate = 0
			@total_dwell_time = 0
			@current_position = Geom::Point3d.new(@env.home_x,@env.home_y,@env.home_z)
			@group_num = 0
			@indentation = 0
		end
		
		attr_accessor :current_position
	
		def set_unit_mode(mode)
			@unit_mode = :inches if mode == :inches
			@unit_mode = :millimeters if mode == :millimeters
		end
		
		def indent
			ret = ""
			ret << "   " if @group_initialized
			ret << "   " if @cut_path_stack_initialized
			ret << "   " if @cut_path_layer_initialized
			return ret
		end
		
		def toggle_spindle(tog)
			@gcode_out << "\n#{indent}M5 (SPINDLE OFF)" if tog == :off
			@gcode_out << "\n#{indent}S16500 M03 (SPINDLE ON)" if tog == :on
		end
		
		def seek_to_safe
			return if @current_position.z == @env.safe_height
			x = @current_position.x
			y = @current_position.y
			new_position = Geom::Point3d.new(x,y,@env.safe_height)
			move_to(new_position, :seek)
		end
		
		def seek_to_material_clearance
			x = @current_position.x
			y = @current_position.y
			new_position = Geom::Point3d.new(x,y,@env.material_clearance)
			move_to(new_position, :seek)
		end
		
		def get_feed_rate(pgl_item)
			z_diff = PathGeomLib::de_float_error((pgl_item.end_pt.z - @current_position.z)).abs
			return @env.feed_rate if z_diff == 0
			horiz_time = pgl_item.length / @env.feed_rate
			vert_rate = z_diff / horiz_time
			if vert_rate > @env.vertical_feed_rate then
				target_horiz_time = z_diff / @env.vertical_feed_rate
				target_horiz_rate = pgl_item.length / target_horiz_time
				target_horiz_rate = @env.vertical_feed_rate if target_horiz_rate < @env.vertical_feed_rate
				return target_horiz_rate
			else
				return @env.feed_rate
			end
		end
		
		def move_to(point, move_type)
			raise "Valid move types are :seek and :feed" unless move_type == :feed or move_type == :seek
			code = ""
			feed_string = ""
			
			seg_st = @current_position.clone
			seg_st.z = point.z
			move_segment = PGL_SegmentDef.new(seg_st,point)
			feed_rate = get_feed_rate(move_segment)
			feed_rate = @env.seek_rate if move_type == :seek
			
			if @current_feed_rate != feed_rate then
				@current_feed_rate = feed_rate
				feed_string = " F%.4f" % [feed_rate]
			end
			
			if move_type == :feed then
				code = "G01"
				@group_feed_distance += @current_position.distance(point) if @group_feed_distance
				@total_feed_distance += @current_position.distance(point)
			else
				code = "G00"
				@group_seek_distance += @current_position.distance(point) if @group_seek_distance
				@total_seek_distance += @current_position.distance(point)
			end
			
			@current_position = point
			@gcode_out <<  "\n#{indent}#{code} #{point_to_xyz(point)}#{feed_string}"
		end
	
		def header
			@gcode_out << "(Generated with Grays42_GPath)\n\n"
			@gcode_out << "(---------- HEADER ----------)\n"
			@gcode_out << "G90 (Absolute distance)\n"
			@gcode_out << "G40 (Tool Radius Compensation Off)\n"
			@gcode_out << "G20 (Inches)\n" if @unit_mode == :inches
			@gcode_out << "G21 (Millimeters)\n" if @unit_mode == :millimeters
			@gcode_out << "G17 (All curves are in X-Y plane)\n"
			@gcode_out << "G94 (Inches per minute feed mode)\n"
			#@gcode_out << "M5 (Start without the motor on)\n"
			@gcode_out << "M03 (Motor on)\n"
			@gcode_out << "\n\n"
			
			@header_is_set = true
			@total_feed_distance = 0
			@total_seek_distance = 0
			@current_feed_rate = 0
			@total_dwell_time = 0
			@current_position = Geom::Point3d.new(@env.home_x,@env.home_y,@env.home_z)
			@group_num = 0
			@indentation = 0
		end
		
		def footer
			
			@gcode_out << "\n\n\n#{indent}(FINAL CLEANUP)"
			seek_to_safe
			
			@gcode_out << "\n\n#{indent}(FINAL WORK RUN REPORT:)"
			@gcode_out << "\n#{indent}(Total feed distance: %.4f #{@unit_mode.to_s}.)" % @total_feed_distance
			@gcode_out << "\n#{indent}(Total seek distance: %.4f #{@unit_mode.to_s}.)" % @total_seek_distance
			@gcode_out << "\n#{indent}(Total dwell time: #{@total_dwell_time} seconds.)" unless @group_dwell_time == 0
			
		end
		
		def cut_group(workpiece)
			initialize_group(workpiece.group)
			workpiece.paths.each do |path|
				cut_path_stack(path)
			end
			close_group
		end
		
		def cut_screw(screw_point)
			seek_to_safe
			
			seek_target = screw_point.clone
			seek_target.z = @env.safe_height
			
			material_target = screw_point.clone
			material_target.z = @env.material_clearance
			
			drill_target = screw_point.clone
			drill_target.z = @env.screw_depth
			
			move_to(seek_target,:seek)
			move_to(material_target,:seek)
			move_to(drill_target,:feed)
			
			seek_to_safe

		end
		
		def initialize_group(object=nil)
			raise "Generate header first" unless @header_is_set
			raise "Group already initialized" if @group_initialized
			@group_initialized = true
			@group_path_stack_num = 0
			
			@group_num = @group_num + 1
			@indentation += 1
			
			@gcode_out << "\n\n#{indent}(---------- WORKPIECE #{@group_num} ----------)"
			if object.class == Sketchup::Group or object.class == Sketchup::ComponentInstance then
				x0 = object.bounds.corner(0).x
				y0 = object.bounds.corner(0).y
				x1 = object.bounds.corner(3).x
				y1 = object.bounds.corner(3).y
				@gcode_out << "\n#{indent}(---------- WORKPIECE NAME: #{object.name} ----------)" unless object.name == ""
				@gcode_out << "\n#{indent}(Group x-y bounds: [#{x0},#{y0}],[#{x1},#{y1}])"
				@gcode_out << "\n#{indent}(Group height: #{object.bounds.corner(7).z})"
				@gcode_out << "\n#{indent}(Group base z: #{object.bounds.corner(0).z})"
			end
			@group_seek_distance = 0
			@group_feed_distance = 0
			@group_dwell_time = 0
			
		end
		
		def close_group
			raise "Initialize group first" unless @group_initialized
			@group_initialized = false
			@group_path_stack_num = 0
			
			@gcode_out << "\n\n#{indent}(WORKPIECE #{@group_num} REPORT:)"
			@gcode_out << "\n#{indent}(Workpiece feed distance: %.4f #{@unit_mode.to_s}.)" % @group_feed_distance
			@gcode_out << "\n#{indent}(Workpiece seek distance: %.4f #{@unit_mode.to_s}.)" % @group_seek_distance
			@gcode_out << "\n#{indent}(Workpiece dwell time: #{@group_dwell_time} seconds.)" unless @group_dwell_time == 0
			@indentation -= 1
		end
		
		def cut_path_stack(path)
			@cut_path_stack_initialized = true
			@gcode_out << "\n\n#{indent}(--- WORKPIECE #{@group_num}, PATH STACK #{@group_path_stack_num} ---)"
			
			#Prep for cut
			seek_to_safe
			start_safe = path.start_pt.clone
			start_safe.z = @env.safe_height
			move_to(start_safe, :seek)
			seek_to_material_clearance
			
			za = @env.cut_z_array
			if @env.gradual_descent then
				cut_path_layer(path,za[0],za[0])
				0.upto(za.length - 2) do |i|
					cut_path_layer(path,za[i],za[i+1])
				end
				cut_path_layer(path,za.last,za.last)
			else
				0.upto(za_length - 1) do |z|
					cut_path_layer(path,z,z)
				end
			end
			@group_path_stack_num += 1
			
			seek_to_safe
			@cut_path_stack_initialized = false
		end
		
		def cut_path_layer(path,start_z,end_z)
			@gcode_out << "\n\n#{indent}(--- WP #{@group_num}, STACK #{@group_path_stack_num},  z#{start_z} -> z#{end_z} ---)"
			@cut_path_layer_initialized = true
			start_pt = path.start_pt.clone
			start_pt.z = start_z
			move_to(start_pt,:feed)
			fed_so_far = 0
			path.pgl_items.each do |item|
				fed_so_far += item.length
				z = start_z - (start_z - end_z)*(fed_so_far/path.length)
				feed_pgl(item.clone,z)
			end
			@cut_path_layer_initialized = false
		end
		
		def feed_pgl(item,z)
			feed_arc(item,z) if item.class == PGL_ArcDef
			feed_segment(item,z) if item.class == PGL_SegmentDef
		end
		
		def feed_arc(arc,z)
			end_pt = arc.end_pt
			end_pt.z = z
			if arc.zero? then
				move_to(end_pt,:feed)
				dwell
			else
				g = "G03"
				g = "G02" if arc.direction == :clockwise
				
				center_offset_vector = arc.start_pt.vector_to(arc.center)	
				@gcode_out << "\n#{indent}#{g} #{point_to_xyz(end_pt)} #{point_to_ij(center_offset_vector)}"
				@current_position = arc.end_pt
				
				@total_feed_distance += arc.length
				@group_feed_distance += arc.length
			end
		end
		
		def feed_segment(seg,z)
			tar_pt = seg.end_pt
			tar_pt.z = z
			move_to(tar_pt,:feed)
			@total_feed_distance += @current_position.distance(seg.end_pt)
			@group_feed_distance += @current_position.distance(seg.end_pt)
		end
		
		def dwell
			@gcode_out << "\n#{indent}G04 P#{@env.dwell_time*1000}"
			@total_dwell_time += @env.dwell_time
			@group_dwell_time += @env.dwell_time
		end
		
		def point_to_xyz(point)
			return "X%.4f Y%.4f Z%.4f" % [point.x, point.y, point.z]
		end
		
		def point_to_ij(vector)
			return "I%.4f J%.4f" % [vector.x, vector.y]
		end

		def gcode
			return @gcode_out.clone
		end
		
	end

end