module Grays42_GPath

	#FLAT GEOMETRY ONLY (XY PLANE)
	#NO Z CALCULATIONS
	
	#The purpose of the PGL library is to fully abstract arcs and lines so they are no longer dependent on the Sketchup edge geometry.
	#This allows for mathematically precise intersections between two arcs or an arc and a line. It also allows the PGL curves and lines to be modified directly without altering sketchup geometry.
	#This prevents imperfect intersections from causing toolpath problems due to the way Sketchup generates curves as "edges that have some extra information".
	#All calculations using this library are guaranteed to be accurate to within 4 significant digits, which is enough for CNC accuracy. Some rounding is needed to match floating point values.
	
	class PGL_Intersect
		attr_accessor :pts
		attr_accessor :separation
		
		def initialize(pts = [], separation = 0)
			@pts = pts
			@separation = 0 #separation by length
		end
		
		def num
			return @pts.length
		end
		
		def none?
			return true if num == 0
			return false
		end
		
	end

	class PathGeomLib
		
		QNAN_STR = "~ -1.#QO\""
		
		class Geom::Point3d
			def contains_qnan?
				[x,y,z].each do |v|
					return true if v.to_s == PathGeomLib::QNAN_STR
				end
				return false
			end
		end
		
		#Needed to correct for floating point errors that cause negative or off-zero results with errors in the -10th magnitude or smaller.
		#The magnitude of my hatred for floating point errors in this code rivals that of the national debt.
		def self.is_near_zero?(val,mag = 4)
			return true if (val*(10**mag)).round == 0
			return false
		end
		
		def self.near_zero(val, mag = 4)
			return 0 if is_near_zero?(val,mag)
			return val
		end
		
		
		
		def self.de_float_error(val, mag=4)
			if val.class == Geom::Vector3d then 
				return Geom::Vector3d.new(de_float_error(val.x),de_float_error(val.y),de_float_error(val.z))
			elsif val.class == Geom::Point3d then
				return Geom::Point3d.new(de_float_error(val.x),de_float_error(val.y),de_float_error(val.z))
			end
			val = 0 if is_near_zero?(val,mag)
			return val
		end
		
		#Returns all intersections between any two items.
		def self.intersect(item1, item2, guaranteed_intersection = false)
			return item1.end_pt if item1.end_pt == item2.start_pt
			
			class1 = item1.class
			class2 = item2.class
			ints = nil
			
			if class1 == PGL_SegmentDef and class2 == PGL_SegmentDef then
				ints = self.intersect_segment_segment(item1, item2)
				raise "seg-seg" unless ints.class == PGL_Intersect
			elsif class1 == PGL_SegmentDef and class2 == PGL_ArcDef then
				ints = self.intersect_segment_arc(item1, item2)
				raise "seg-arc" unless ints.class == PGL_Intersect
			elsif class1 == PGL_ArcDef and class2 == PGL_SegmentDef then
				ints = self.intersect_arc_segment(item1, item2)
				raise "arc-seg" unless ints.class == PGL_Intersect
			elsif class1 == PGL_ArcDef and class2 == PGL_ArcDef then
				ints = self.intersect_arc_arc(item1, item2)
				raise "arc-arc" unless ints.class == PGL_Intersect
			else
				raise "Error: no valid intersection types"
			end
			
			return nil if ints.nil?
			
			int = process_intersect(item1,item2,ints, guaranteed_intersection)
			return int.pts[0]
		end
		
		def self.intersect_arc_segment(arc,segment)
			return intersect_line_circle(segment,arc)
		end
		
		def self.intersect_segment_arc(segment, arc)
			return intersect_line_circle(segment,arc)
		end
		
		def self.intersect_arc_arc(arc1,arc2)
			if arc1.circle? and arc2.circle? and arc1 == arc2 then
				return PGL_Intersect.new([arc1.start_pt])
			elsif arc1.center == arc2.center and arc1.radius != arc2.radius then
				return PGL_Intersect.new
			elsif arc1.center == arc2.center and arc1.radius == arc2.radius then
				return PGL_Intersect.new([arc2.start_pt]) if arc1.distance_from(arc2.start_pt) == 0
				raise "Error: have to fix this, two arcs overlap and don't share an endpoint, not sure how"
			end
			return intersect_circle_circle(arc1,arc2)
		end
		
		def self.intersect_segment_segment(seg1,seg2)
			line_intersection = intersect_line_line(seg1,seg2)
		
			#First, address the case that two segments exist on the same line.
			if line_intersection.separation == :line then
				
				#Solve by sorting points per suggestion here: http://stackoverflow.com/a/17152247				
				ref_pts = []
				[seg1,seg2].each do |seg|
					seg.points.each do |pt|
						ref_pts << pt unless ref_pts.include?(pt)
					end
				end
				
				#Note: duplicate points have been overwritten; the joint of two perfect endpoints
				#is only one pt, for example.
				
				#Now sort the hash so that we can pick our two middle points.
				if seg1.m_infinity? then
					#Case 1: line is vertical. Order by y points because ordering by x doesn't work.
					ref_pts.sort {|p1,p2| p1.y <=> p2.y}
				else
					#Case 2: all other slopes, Order by x poincts.
					ref_pts.sort {|p1,p2| p1.x <=> p2.x}
				end
				
				#Case: if we have 4 points, then we are interested in the two middle points.
				if ref_pts.length == 4 then
					#If both lines include both middle points, this is an overlap.
					
					#Both lines will include both points iff either line includes both points due to constraints of method.
					if seg1.distance_from(ref_pts[1]) == 0 and seg1.distance_from(ref_pts[2]) == 0 then
						#EXECUTIVE DECISION: In the case of all overlaps (which should never happen anyway),
						#we do not care about any of the points. We just want an intersection. The first point encountered will suffice.
						return PGL_Intersect.new([ref_pts[1]])
					else
						#If they don't include both, this is not an overlap; the two middle points do not connect.
						#Segments can ONLY intersect at precise points.
						return PGL_Intersect.new
					end
					
				#Case: if we have 3 points, the middle point is the joint of two segments.
				elsif ref_pts.length == 3 then
					#Make sure the lines include it.
					return PGL_Intersect.new if de_float_error(seg1.distance_from(ref_pts[1])) > 0 or de_float_error(seg2.distance_from(ref_pts[1])) > 0
					return PGL_Intersect.new([ref_pts[1]])
				
				#Case: if we have 2 points, then the two lines were identical. Pick either point.
				elsif ref_pts.length == 2 then
					return PGL_Intersect.new([ref_pts[0]])
				#Case: anything else is impossible.
				else
					return PGL_Intersect.new([ref_pts[0]])
				end
			else
				line_intersection.pts.each do |pt|
					line_intersection.pts.delete(pt) if seg1.distance_from(pt) > 0 or seg2.distance_from(pt) > 0
				end
				return line_intersection
			end
		end
		
		#This finds the intersects of two LINES, not line SEGMENTS.
		#This will find the intersect point whether it's included in the line segment or not!
		def self.intersect_line_line(line1,line2)
			#intersection for y and x for equations:
			# y = m1*x + b1
			# y = m2*x + b2
			
			m1 = line1.m
			m2 = line2.m
			b1 = line1.b
			b2 = line2.b

			#Case: identical lines. Also captures verticals.
			if m1.class == m2.class and m1 == m2 then
				#Store this in a (TEMPORARY) intersect container (kinda hacky because this isn't really an intersect) to send to segment resolution.
				ret = PGL_Intersect.new([line1,line2])
				ret.separation = :line
				return ret
			end
			
			#All other cases can be determined by Sketchup.
			p1 = line1.start_pt
			v1 = p1.vector_to(line1.end_pt)
			p2 = line2.start_pt
			v2 = p2.vector_to(line2.end_pt)
			
			#If either vector is nil, then we have a point and not a line. Intersect only occurs if that point is ON the other line/point.
			#Otherwise, just connect the lines.
			null_vec = Geom::Vector3d.new
			intersect = nil
			if v1 == null_vec and v2 == null_vec then
				intersect = p1 if p1 == p2
			elsif v1 == null_vec then
				intersect = p1 if p1.on_line?([p2,v2])
			elsif v2 == null_vec then
				intersect = p2 if p2.on_line?([p1,v1])
			else
				intersect = Geom.intersect_line_line([p1,v1],[p2,v2])
			end
			
			if intersect.nil? then
				return PGL_Intersect.new
			else
				return PGL_Intersect.new([intersect])
			end
		end
		
		
		
		#This will determine the intersection points for the circles described by two PGL arcs. This WILL NOT consider whether the circles intersect but the arcs don't.
		#In other words, an extra step will be needed to determine whether the arcs actually intersect.
		def self.intersect_circle_circle(circle1,circle2)
			#Using a pure geometrical solution found here:
			#http://www.sonoma.edu/users/w/wilsonst/papers/geometry/circles/T11-15/T13.html
			#For each equation, breaking the area before the plus/minus into "part1x" and "part1y", the thing after "part2x" and "part2y".
			#The part under the square root is "part_sqrt", and determines whether there are two solutions, one solution, or no solutions
			
			r1 = circle1.radius.abs.to_l
			r2 = circle2.radius.abs.to_l
			x1 = circle1.center.x.to_l
			x2 = circle2.center.x.to_l
			y1 = circle1.center.y.to_l
			y2 = circle2.center.y.to_l
			z = circle1.center.z
			
			#distance between centers
			d = (( (x2 - x1)**2 + (y2 - y1)**2 )**(1.0/2)).to_l
			raise "Error: circles are identical, filter prior to intersect_circle_circle" if d == 0
			
			#automatically rule out intersection if the centers are just too far from each other
			#Note: account for exactly the right distance (not sure if this will throw floating point errors)
			return PGL_Intersect.new if d > r1 + r2 and not r1 + r2 == d

			#NOTE: If I ever have to make this a length for rounding, AMPLIFY BEFORE part_sqrt == 0 check below or
			# will throw errors on certain very small precision joins
			part_sqrt = (((r1 + r2)**2 - d**2)*(d**2-(r1-r2)**2))
			return PGL_Intersect.new if part_sqrt < 0
			
			part1x = (x2 + x1)/2 + (x2 - x1)*(r1**2 - r2**2)/(2*(d**2))
			part2x = (y2 - y1)*(part_sqrt**(1.0/2))/(2*(d**2))
			
			part1y = (y2 + y1)/2 + (y2 - y1)*(r1**2 - r2**2)/(2*(d**2))
			part2y = -(x2 - x1)*(part_sqrt**(1.0/2))/(2*(d**2))

			p1 = Geom::Point3d.new(part1x + part2x, part1y + part2y, z)
			p2 = Geom::Point3d.new(part1x - part2x, part1y - part2y, z)
			
			return PGL_Intersect.new([p1,p2])
		end
		
		#This combines all final processing steps prior to releasing the intersect back into path handling routines.
		def self.process_intersect(item1,item2,intersect, guaranteed_intersection)
			check_tolerance(item1,item2,intersect)
			check_directionality(item1,item2,intersect)
			closest_intersect(item1,item2,intersect)
			return intersect
		end
		
		def self.check_directionality(item1,item2,intersect)
			return		
			
			i1p = [item1.start_pt,item1.end_pt]
			i2p = [item2.start_pt,item2.end_pt]
			intersect.pts.each do |pt|
				next unless i1p.include?(pt) and i2p.include?(pt)
				intersect.pts.delete(pt) if pt == item1.start_pt and (pt == item2.end_pt or pt == item2.start_pt)
				intersect.pts.delete(pt) if pt == item2.end_pt and (pt == item2.end_pt or pt == item2.start_pt)
			end
		end
		
		def self.is_precise?(item1,item2,intersect)
			return false if intersect.nil?
			return true if item1.distance_from(intersect) == 0 and item2.distance_from(intersect) == 0
			return false
		end
		
		def self.check_tolerance(item1,item2,intersect)
			return false if intersect.nil?
			delete_pts = []
			intersect.pts.each do |pt|
				delete_pts << pt unless item1.within_separation_tolerance(1,pt)
				delete_pts << pt unless item2.within_separation_tolerance(0,pt)
			end
			delete_pts.each {|d| intersect.pts.delete(d)}
		end
		
		#This checks the distance between the segments and the intersect points and trims away all but the closest intersect(s).
		#It also sets separation distance.
		def self.closest_intersect(item1,item2,intersect)
			return false if intersect.nil?
			case intersect.pts.length
				when 1
					intersect.separation = item1.distance_from(intersect.pts[0]) + item2.distance_from(intersect.pts[0])
					return
				when 2
					dist1 = item1.distance_from(intersect.pts[0]) + item2.distance_from(intersect.pts[0])
					dist2 = item1.distance_from(intersect.pts[1]) + item2.distance_from(intersect.pts[1])
					
					if dist1 < dist2 then
						intersect.pts.delete_at(1)
						intersect.separation = dist1
					elsif dist1 > dist2 then
						intersect.pts.delete_at(0)
						intersect.separation = dist2
					else
						intersect.separation = dist1
					end
				end
			return intersect
		end
		
		def self.intersect_line_circle(line, circle)
			#A circle is described in standard form as (X - c.x)^2 + (Y - c.y)^2 = r^2
			#Straight line described by y = mx + b
						
			z = circle.center.z
			
			#Case: m is infinite, which means line is vertical
			if line.m_infinity? then
				#(p1.x - c.x) -> x_term					
				#x_term + (Y - c.y) = r
				#Y - c.y = +/- sqrt(r- x_term)
				#sqrt(r - x_term) -> var_term
				#Y = (+/- var_term) + c.yx
				x = line.points[0].x
				x_term = ((x - circle.center.x)**2).to_l
				
				var_term = (circle.radius**2 - x_term).to_l
				var_term = 0 if var_term == 0  #Necessary because taking the square root of -0.000 still throws an error
				
				return PGL_Intersect.new if var_term < 0
				sq_term = Math.sqrt(var_term).to_l
				
				y1 = sq_term + circle.center.y
				y2 = -sq_term + circle.center.y
				
				p1 = Geom::Point3d.new(x,y1,z)
				p2 = Geom::Point3d.new(x,y2,z)
				
				raise "Error: invalid point p1 #{p1}" if p1.contains_qnan?
				raise "Error: invalid point p2 #{p2}" if p2.contains_qnan?
				
				return PGL_Intersect.new([p1]) if y1 == y2
				return PGL_Intersect.new([p1,p2])
			end
			
			#Long form algebraic solution (worked out by hand) for intersection between
			#circle (X - h)^2 + (Y - k)^2 = r^2 and line Y = mX + b
			#0 = X(1 + m) + X(-2h + 2mb - 2mk) + (h - 2bk + b + k - r)
			
			h = circle.center.x
			k = circle.center.y
			m = line.m
			b = line.b
			r = circle.radius.abs.to_l
			
			qa = (1 + m**2).to_l
			qb = (-2*h + 2*m*b - 2*m*k).to_l
			qc = (h**2 - 2*b*k + b**2 + k**2 - r**2).to_l

			disc = (qb**2 - 4.0*qa*qc).to_l
			disc = 0.to_l if disc == 0   #Necessary because disc -0.000 still throws errors when square root is taken
			
			return PGL_Intersect.new if disc < 0
			
			x_pos = ((-qb + ((disc)**(1.0/2.0)))/(2*qa)).to_l
			x_neg = ((-qb - ((disc)**(1.0/2.0)))/(2*qa)).to_l
			
			y_pos = m*x_pos + b
			y_neg = m*x_neg + b
			
			
			
			p1 = Geom::Point3d.new(x_pos,y_pos,z)
			p2 = Geom::Point3d.new(x_neg,y_neg,z)
			
			raise "Error: invalid point p1 #{p1}" if p1.contains_qnan?
			raise "Error: invalid point p2 #{p2}" if p2.contains_qnan?
			
			return PGL_Intersect.new([p1,p2])
		end
		
	end
	

	

end