module Grays42_GPath
	
	#PGL_ArcDef is the mathematical representation of an arc segment for the purpose of intersections and other calculations.
	
	class PGL_ArcDef < PGL_Item
		
		#The only purpose of the normal is to determine where the arc sweeps (i.e. directionality)
		#A positive normal Z is COUNTER-CLOCKWISE. A negative normal Z is CLOCKWISE.
		#The deviation is the maximum distance that two intersections can be "off" to make a valid intersection. This is used to correct for imperfections
		#in the rendering of the curve. If separation allowed is 0, the intersection must be precise. In practice, the maximum deviation of any intersection
		# is the length of one of the edges that comprises the arc in Sketchup geometry.
		def initialize(center, radius, start_vector,end_vector,normal = Geom::Vector3d.new(0,0,1),separation_allowed = 0)
			@center = center
			@radius = radius.to_l
			@normal = normal
			@zero = false  #Note that this should only be called in the event that the vectors are the same (i.e. is either a full circle or nothing)
			@separation_allowed = separation_allowed
			
			#Vectors are ordered, and scalars. Length is always 1.
			@vectors = [start_vector.normalize,end_vector.normalize]
			@sweep = sweep_between_vectors(@vectors[0],@vectors[1])
		end
		
		attr_accessor :separation_allowed
		
		def radius
			return @radius
		end
		
		def radius=(radius)
			@radius = radius.to_l
		end
		
		#When two vectors are equal, we don't know whether we're working with a full circle or an arc with no length.
		#Code that creates or modifies arcs is responsible for flagging zeroed-out curves correctly.
		#Otherwise, the asusmption is that an arc with two identical vectors IS A FULL CIRCLE.
		def zero!
			raise "Error: call zero! only when distinguishing between a circle or a zeroed curve." unless @vectors[1] == @vectors[0]
			@zero = true
			@sweep = 0
		end
		
		def zero?
			return true if @radius == 0
			return true if @zero
			return false
		end
		
		def offset!(offset_amnt)
			# We always offset to the LEFT of the direction of travel (if positive amount)
			# If we're going clockwise, increase the radius. If we're going counterclockwise, decrease the radius.
			multiplier = (@normal.normalize).z
			@radius = (@radius - offset_amnt*multiplier).to_l
		end
		
		def transform!(transformation)
			@center.transform!(transformation)
			@normal.transform!(transformation)
			@vectors.each do |v|
				v.transform!(transformation)
			end
		end
		
		def ==(other_item)
			return false unless other_item.class == PGL_ArcDef
			return false unless @center == other_item.center
			return false unless (@radius*10**10).round == (other_item.radius*10**10).round
			return false unless @vectors[0] == other_item.vectors[0]
			return false unless @vectors[1] == other_item.vectors[1]
			return false if @zero ^ other_item.zero?
			return true
		end
		
		def describe
			rotation = "CCW"
			rotation = "CW" if @normal.z < 0
			return "#{rotation} arc (#{start_pt.x},#{start_pt.y})->(#{end_pt.x},#{end_pt.y}), c=#{@center.x},#{@center.y}, r=#{@radius}, sweep=#{@sweep*180.0/Math::PI} deg, sep #{@separation_allowed}\n"
		end
		
		def draw(entities = Sketchup.active_model.entities)
			entities.add_arc(*draw_vars)
		end
		
		def offset(amnt)
			ret = clone
			ret.offset!(amnt)
			return ret
		end
		
		def direction
			return :clockwise if @normal.z < 0
			return :counter_clockwise if @normal.z > 0
			return nil
		end
		
		def circle?
			return true if @vectors[0] == @vectors[1] and not zero?
			return false
		end
		
		def normalize!
			@vectors.each {|v| v.normalize! }
			@normal.normalize!
		end
		
		def vectors
			return [@vectors[0].clone,@vectors[1].clone]
		end
		
		attr_accessor :radius, :center
		
		def normal
			return @normal.clone
		end
		
		def sweep
			return @sweep
		end
		
		def tangent(index)
			return @vectors[index].cross(@normal.reverse).normalize
		end
		
		#Splits the two. Edits own values for the first half, returns the second half.
		def split(point)
			raise "Error: cannot split PGL item using a point that is not on the item." unless distance_from(point) == 0
			second = clone
			second.set_pt(0,point)
			set_pt(1,point)
			return second
		end
		
		def get_halves(point)
			item1 = clone
			item2 = item1.split(point)
			return [item1,item2]
		end
		
		#Note: unless forced, this only *approximates* the position denoted by a point. It should match ok with a precise point, but if it's
		#given a point not actually on the arc it will still attempt to do it even though it'll be slightly off. This can happen if
		#the intercept point is partway through an "edge" on the Sketchup geometry "curve", in which case the intersection won't
		#be the same as a true curve. This is handled properly when the geometry is bound together in a PGL_Path.
		def set_pt(index,pt)
			puts "#{@center}, #{pt.x.to_s == "~ -1.#QO\""}, #{@center.distance(pt)} == #{@radius} #{@radius.class}" unless @center.distance(pt) == @radius.abs.to_l
			raise "Error: attempted to set endpoint at an invalid point" unless @center.distance(pt) == @radius.abs.to_l
			vec = @center.vector_to(pt)
			vec.reverse! if @radius < 0
			set_vector(index,vec)
		end
		
		def set_vector(index,vector)
			@vectors[index] = vector.normalize
			@zero = false unless @vectors[1] == @vectors[0]
			@sweep = sweep_between_vectors(@vectors[0],@vectors[1])
		end
		
		def start_vector
			ret = @vectors[0].clone
			ret.length = radius
			return ret
		end
		
		def end_vector
			ret = @vectors[1].clone
			ret.length = radius
			return ret
		end
		
		def is_same_circle?(other_arc)
			return false unless other_arc.class == PGL_ArcDef
			return true if @center == other_arc.center and @radius == other_arc.radius
			return false
		end
		
		def within_separation_tolerance(reference_index,pt)
			return true if distance_from(pt) == 0
			
			reference = end_pt if reference_index == 1
			reference = start_pt if reference_index == 0
			
			dist = reference.distance(pt)
			return true if dist<= @separation_allowed
			return false
		end
		
		#Fuses two pgl items together. If it's a continuation of the same arc, steal the endpoint and zero the arc.
		def fuse!(other_item,guaranteed_intersection = false)
			int_pt = intersect(other_item, guaranteed_intersection)
			return false if int_pt.nil?
			if other_item.class == PGL_ArcDef then
				if other_item == self then
					can_join_with_self = start_pt.distance(end_pt) <= @separation_allowed
					#Apparently this isn't necessary, can_join_with_self currently only occurs in the event of a full circle that already has this pt set anyway
					#set_pt(1,start_pt) if can_join_with_self
					return true if can_join_with_self
					return false
				else
					set_pt(1,other_item.end_pt) if is_same_circle?(other_item)
					other_item.set_pt(0,other_item.end_pt) if is_same_circle?(other_item)
					other_item.zero! if is_same_circle?(other_item)
					return true
				end
			else
				set_pt(1,int_pt)
				other_item.set_pt(0,int_pt)
				return true
			end
		end
		
		def clone
			return PGL_ArcDef.new(@center.clone, @radius, start_vector.clone, end_vector.clone, @normal.clone)
		end
		
		def intersect(other_item, guaranteed_intersection = false)
			return PathGeomLib.intersect(self, other_item, guaranteed_intersection)
		end
		
		def reverse!
			@normal.reverse!
			@vectors.reverse!
		end
		
		def reverse
			ret = clone
			ret.reverse!
			return ret
		end
		
		def sweep_between_vectors(v1,v2)
			dot = v1.dot(v2)
			det = @normal.dot(v1.cross(v2))
			
			result = Math::atan2(det,dot) % (Math::PI*2.0)
			begin
				result = Math::PI*2.0 if result == 0    #If we came up with zero, assume we have a full circle.
			rescue
				#puts "WARNING: FIX THIS"
			end
			return result
		end
		
		#This determines the angular distance from the point to the arc's sweep.
		#If the point is in the arc's sweep, return will be 0.
		#If the point is not in the arc's sweep, return will be the minumum angular distance the point is away from one of the two vectors.
		#The purpose of this function is to determine which of two intersections is "closest" to the arc to correct for Sketchup arc geometry problems.
		def angular_distance_from(point)
			#This should never really happen but it's technically in the inclusive sweep and I don't want it throwing zero-length vector errors
			return 0 if point == @center
			
			raise "Can only test point on same plane as arc" unless point.z == @center.z
			test_vector = @center.vector_to(point).normalize
			test_sweep = sweep_between_vectors(@vectors[0],test_vector)
			return 0 if test_sweep <= @sweep
			
			#(Working this out in my head)
			#half_of_negative_sweep = (360 - sweep) / 2
			#sweep_error = test_sweep - sweep tells me how far into the negative sweep I am.
			#if sweep_error > half_of_negative_sweep then return negative sweep - sweep_error
			sweep_error = test_sweep - @sweep
			negative_sweep = Math::PI*2.0 - @sweep
			if sweep_error < negative_sweep/2.0 then
				return sweep_error
			else
				return negative_sweep - sweep_error
			end
		end
		
		#This determines the actual measured distance between any point and the arc.
		def distance_from(point)
			angular_dist = angular_distance_from(point)
			#If angular distance is 0, then the point is inside the arc's sweep and the distance is simply how far away from the radius it is.
			return ((@center.distance(point) - @radius).abs).to_l if angular_dist == 0
			#Otherwise the closest point will be one of the two endpoints.
			distance = [end_pt.distance(point),start_pt.distance(point)].min
			return distance
		end
		
		def start_pt
			pt = @center.clone
			pt.z = get_z if z?
			return pt if @radius == 0
			vec = @vectors[0].normalize.clone
			vec.length = @radius
			return pt.offset(vec)
		end
		
		def end_pt
			pt = @center.clone
			pt.z = get_z if z?
			return pt if @radius == 0
			vec = @vectors[1].normalize.clone
			vec.length = @radius
			return pt.offset(vec)
		end
		
		def draw_vars(sides = 12)
			return [@center.clone,@vectors[0].clone,@normal.clone,@radius,0.0,@sweep]
		end
		
		def length
			return sweep*radius 
		end
		
=begin
		
		#Returns all the variables necessary to draw the curve in Sketchup
		# center, xaxis, normal, radius, start_angle, end_angle, num_segments
		# Passing the result with the splat (*) operator will drop them in nicely. Hopefully. This is the first time I've used splat like this.
		def draw_vars(num_segments = 12)
			xaxis = @start_vector
			end_angle = arc_sweep
			end_angle = -arc_sweep if @is_clockwise
			
			return [@center, xaxis, @plane_normal, @radius, 0.0, end_angle, num_segments]
		end

		def tangent_at(start_or_end)
			eval_vector = nil
			cross_vector = @plane_normal.clone
			if start_or_end == :start then
				eval_vector = @start_vector.reverse
			elsif start_or_end == :end then
				eval_vector = @end_vector.clone
			else
				return nil
			end
			
			
			tangent_vector = eval_vector.cross(cross_vector).normalize
			tangent_vector.reverse! if @is_clockwise
			
			return tangent_vector
		end
		
#		def is_point?
#			return true if @start_vector.length == 0
#		end
		
#		def is_negative?
#			return true if @radius < 0
#		end
		
#		def is_clockwise
#			return @is_clockwise
#		end
		
#		def is_full_circle
#			return true if (arc_sweep*10000.0).round == (2*Math::PI*10000.0).round
#			return false
#		end
		
		def include_point?(p)
			#Step 1: is it even on the line?
			#circle (X - h)^2 + (Y - k)^2 = r^2
			result = (p.x - @center.x)**2 + (p.y - @center.y)**2 - @radius**2
			return false if result.nan?
			return false unless (result*10000.0).round == 0
			
			
			total_angle = arc_sweep
			
			pvec = @center.vector_to(p)
			
			#If the vector is equal to or between the start and end vectors, then the angles between the vector and start/end vectors will sum to the arc sweep.
			#If the vector is outside this range, it will be far larger, as it will swing all the way around to match.
			
			angle1 = angle_between(@start_vector, pvec)
			angle1 = 0.0 if (angle1*100000).round == 0 or (angle1*100000).round == ((2*Math::PI)*100000).round
			angle2 = angle_between(pvec,@end_vector)
			angle2 = 0.0 if (angle2*100000).round == 0 or (angle2*100000).round == ((2*Math::PI)*100000).round
			
			if ((total_angle - (angle1 + angle2))*100000).round == 0 then
				return true
			end
			
			return false
		end
=end

	end
end