module Grays42_GPath

=begin
	This is basically an abstract class to make sure I have everything covered for each item.
	All of the following methods must be consistently implemented on arcs, segments, paths. In the future, I may add beziers/cubics or regular curves.
	
	Definition and purpose of a PGL item:
	 A PGL (PathGeomLib) item is an abstract mathematical representation of a directional geometric entity for the purpose of converting
	 to a precise CNC path. PGL items can be arcs, segments, paths, and eventually other items like beziers and cubics. The purpose of 
	 PGL items are to correct Sketchup's horrible implentation of arcs; in Sketchup, arcs are essentially a collection of edges
	 with some curvature information, which causes a mess any time arcs intersect, edges intersect arcs, etc.
	
	For example, take an awkward 13-sided arc (for example). Intersect a line somewhere at an arbitrary point in the middle of one of the
	edges of the curve; while the arc will "theoretically" intersect at a nice clean position, in actuality the intersection point (edge-edge)
	isn't actually on the curvature line. The purpose of pure mathematical PGL items is to eliminate that problem by abstracting
	 out all of the geometry to be manipulated, smooth, clean paths that can be stable and consistent.
	
	Most importantly, all PGL items bake in the ability to offset to the left (positive) or right (negative) of the direction of
	travel. Individual PGL items can accomplish this easily, but the real power comes in PGL_Path, which will automatically trim lines and
	build corners as needed.
=end
	class PGL_Item
		
		def raise_pgl_item_error
			raise "This method must be implemented in class #{self.class}."
		end
		
		#PGL_Items are always directional, this identifies the current start of the item.
		def start_pt
			raise_pgl_item_error
		end
		
		def end_pt
			raise_pgl_item_error
		end
		
		#This applies a standard Sketchup transformation to the item
		def transform!(transformation)
			raise_pgl_item_error
		end
		
		def set_z(z)
			@z = z
		end
		
		def z?
			return true if @z
			return false
		end
		
		def get_z
			return @z
		end
		
		#Replies with true if the PGL item is zero length or less. (Negative curves, for example.)
		#It is the responsibility of the parent of the PGL item to dispose of the item if it is zero status.
		#Normal operations executed on a zero item may not work properly.
		def zero?
			raise_pgl_item_error
		end
		
		def length
			raise_pgl_item_error
		end
		
		#Creates a deep copy
		def clone
			raise_pgl_item_error
		end
		
		#Applies a path reversal; directionality of all PGL components involved in the item are swapped.
		def reverse!
			raise_pgl_item_error
		end
		
		#Returns a cloned reverse path
		def reverse
			raise_pgl_item_error
		end
		
		#This provides one or more intersection points between two PGL items, and specifically returns the PGL_Intersect type.
		def intersect(other_item)
			raise_pgl_item_error
		end
		
		#Attempts to join to the given item. In the event that they represent the same segment or arc, the calling item will "steal" the endpoint
		#of the second item and zero it out, effectively taking it over. Functions that call fuse! should check to see if the second item is zeroed.
		def fuse!(other_item)
			raise_pgl_item_error
		end
		
		#Provides the distance away from the point the PGL is. Useful for determining whether a point is on the PGL item or which of
		#several points is closest to it for intersection purposes.
		def distance_from(point)
			raise_pgl_item_error
		end
		
		#Offsets the path in the given direction (according to directionality). Must handle all cleanup.
		#This returns a copy of the item.
		# NOTE: Offset is always to the LEFT of the current direction of travel when the amount is positive. This applies for all normals too!
		def offset(amnt)
			raise_pgl_item_error
		end
		
		#Offsets the item itself.
		def offset!(amnt)
			raise_pgl_item_error
		end
		
		#Gives the normalized tangent vector at a given index endpoint (0=start,1=end), IN THE DIRECTION OF TRAVEL.
		#For example, the tangent vector at both the start and the end of a segment will be in the direction of the segment.
		def tangent(index)
			raise_pgl_item_error
		end
		
		#Returns a plain-text description of the entity (as a debug tool):
		def describe
			raise_pgl_item_error
		end
	end

end