module Grays42_GPath

	#This class is a way to attach together PGL items in a manageable way in order to execute mass operations cleanly.
	#Note: this CANNOT be given a pile of PGL items and have it sort out what goes where. PGL_Path is the container and items
	#must be given to it in structured order with APPROXIMATELY joining endpoints. (Geometry of PGL items with "theoretically"
	#matching endpoints will automatically be fused in the correct locations, to accomodate Sketchup's horrible curve implementation.)
	#It is the responsibility of other code to figure out how to sort it out.

	class PGL_Path < PGL_Item
		
		def initialize(array = [])
			@pgl_items = []
			@closed = false
			@start_pt = nil
			array.each do |a|
				self << a
			end
		end
		
		attr_accessor :pgl_items
		
		#The function "length" is taken over by actual path length, not array length, so I need a different method for this.
		def array_length
			return @pgl_items.length
		end
		
		def de_float_error(val)
			return PathGeomLib.de_float_error(val)
		end
		
		def <<(pgl_item)
			append_item(pgl_item)
		end
		
		def circle?
			return false unless @pgl_items.length == 1
			return false unless @pgl_items[0].class == PGL_ArcDef
			return false unless @pgl_items[0].start_vector == @pgl_items[0].end_vector
			return true
		end
		
		def append_item(pgl_item, corner_amount = nil, guaranteed_intersection = false)
			raise "Error: cannot append items to a closed path" if @closed
			if @pgl_items.length == 0 then
				@pgl_items << pgl_item
				@start_pt = @pgl_items.first.start_pt
			else
				last_item = @pgl_items.last
				fuse_bool = last_item.fuse!(pgl_item, guaranteed_intersection)
				return false if corner_amount == nil and fuse_bool == false
				corner = build_adjoining_corner(@pgl_items.last,pgl_item,corner_amount)
				@pgl_items << corner if fuse_bool == false and not corner.nil?
				#have to check pgl_item == @pgl_items.last due to 1-item-path circles checking against themselves during the close routine
				@pgl_items << pgl_item unless pgl_item == @pgl_items.last or pgl_item.zero?
			end
			return true
		end
		
		def describe
			ret = "PGL_Path description:\n"
			i = 0
			@pgl_items.each do |item|
				ret << "#{i}: #{item.describe}"
				i = i+1
			end
			return ret
		end
		
		def transform(transformation)
			return clone.transform!(transformation)
		end
		
		def transform!(transformation)
			@pgl_items.each do |e|
				e.transform!(transformation)
			end
		end
		
		def ==(other_path)
			return false unless other_path.is_a?(PGL_Path)
			return false unless other_path.array_length == array_length
			0.upto(@pgl_items.length - 1) do |i|
				return false unless @pgl_items[i] == other_path.pgl_items[i]
			end
			return true
		end
		
		def first
			return @pgl_items.first
		end
		
		def last
			return @pgl_items.last
		end
		
		def fuse!(other_item)
			raise "Path fusions are absurdly complicated and not yet implemented. Also, probably unnecessary. Maybe later."
		end
	
		def closed?
			return @closed
		end
		
		def draw(ents = Sketchup.active_model.entities,separate_operation = false)
			Sketchup.active_model.start_operation "Building path" if separate_operation
			group = ents.add_group
			@pgl_items.each {|item| item.draw(group.entities) }
			Sketchup.active_model.commit_operation if separate_operation
			return group
		end
		
		def verify!
			0.upto(@pgl_items.length - 2) do |i|
				raise "Mismatched endpoints between path items #{i} (#{@pgl_items[i].class}) and #{i+1} #{@pgl_items[i+1].class})" unless @pgl_items[i].end_pt == @pgl_items[i+1].start_pt
			end
			raise "Start/end points do not match on closed loop." if closed? and start_pt != end_pt
		end
		
		def start_pt
			#added a separate start_pt because if the starting PGL gets zeroed (like a corner pocket arc for a joining wood piece that's supposed to zero),
			#then the starting PGL gets removed and the corner pocket arc never completes due to the start point moving.
			return @start_pt
			return nil if @pgl_items.length == 0
			return @pgl_items.first.start_pt
		end
		
		def end_pt
			return nil if @pgl_items.length == 0
			return @pgl_items.last.end_pt
		end
		
		#Replies with true if the PGL item is zero length or less. (Negative curves, for example.)
		#It is the responsibility of the parent of the PGL item to dispose of the item if it is zero status.
		#Normal operations executed on a zero item may not work properly.
		def zero?
			return true if length == 0
		end
		
		def length
			sum = 0
			@pgl_items.each do |item|
				sum = sum + item.length
			end
			return sum
		end
		
		def clone
			new_path = PGL_Path.new
			@pgl_items.each do |item|
				new_path.append_item(item.clone)
			end
			new_path.close! if @closed
			return new_path
		end
		
		#Applies a path reversal; directionality of all PGL components involved in the item are swapped.
		def reverse!
			@pgl_items.each do |item|
				item.reverse!
			end
			@pgl_items.reverse!
		end
		
		#Returns a cloned reverse path
		def reverse
			ret = clone
			ret.reverse!
			return ret
		end
		
		def close!(corner_amount = nil)
			return if array_length == 0
			if start_pt == end_pt then
				@closed = true
				return
			end
			@closed = false
			append_item(@pgl_items.first,corner_amount)
			@pgl_items.pop unless @pgl_items.length == 1
			@closed = true
			set_array_precise
			final_pass
			raise "Error: could not close because the final PGL item is not joined properly." unless @pgl_items.first.start_pt == @pgl_items.last.end_pt
			@start_pt = @pgl_items.first.start_pt
		end
		
		#Offsets the path in the given direction (according to directionality). Must handle all cleanup.
		#Remember, offsets in POSITIVE amounts offset to the LEFT of the path direction. NEGATIVE offsets offset to the RIGHT of the path. (Positive normal.)
		#Builds a new array from scratch, and at each piece checks for intersects and snips off excess, also builds corners where needed.
		def offset!(amnt)
			return false if zero?
			return false if amnt == 0
			new_pgl = PGL_Path.new
			
			0.upto(@pgl_items.length - 1) do |i|
				item = @pgl_items[i].offset(amnt)
				new_pgl.append_item(item,amnt,false)
			end
			
			#If the first item comes out to be zero, this causes an error in the close! routine, so I'm cleaning it up here.
			while new_pgl.array_length > 0 and new_pgl.pgl_items.first.zero?  do
				new_pgl.pgl_items.shift
			end
			
			@pgl_items = new_pgl.pgl_items
			@start_pt = new_pgl.start_pt
			close!(amnt) if @closed
			final_pass
			
			return true
		end
		
		#Break all items at their intersections
		#CURRENTLY FUNCTIONAL AND TESTED BUT UNUSED
		def fracture!
			@pgl_items.each do |item1|
				j_index = @pgl_items[@pgl_items.index(item1)+1 .. @pgl_items.length - 1]
				j_index.each do |item2|
					int = item1.intersect(item2)
					if int then
						
						item1a = item1.split(int)
						@pgl_items.insert(@pgl_items.index(item1)+1,item1a) if item1a
						
						item2a = item2.split(int)
						@pgl_items.insert(@pgl_items.index(item2)+1,item2a) if item2a
						j_index.insert(j_index.index(item2)+1,item2a) if item2a
						
					end
				end
			end
		end
		
		#CURRENTLY FUNCTIONAL AND TESTED BUT UNUSED
		def mend!
			i = 0
			while i < @pgl_items.length do
				j = (i+1) % @pgl_items.length
				fuse_success = @pgl_items[i].fuse!(@pgl_items[j])
				if fuse_success and @pgl_items[j].zero? then
					@pgl_items.delete_at(j)
					i = i - 1
				end				
				i = i + 1
			end
		end
		
		def offset(amnt)
			ret = clone
			ret.offset!(amnt)
			return ret
		end
		
		#This has a single purpose and very specific inputs: builds a corner between two identical-offset pgl items that no longer connect.
		#It will break if it is not given precisely accurate conditions:
		#  1. Corner to build must have a sweep of (0,180]
		#  2. Endpoints must be equidistant from corner center
		#  3. Paths must be going in the same direction with respect to the corner.
		def build_adjoining_corner(item1,item2, offset_amnt)
			return nil if offset_amnt == nil
			p1 = item1.end_pt
			#t1 = item1.tangent(1)
			p2 = item2.start_pt
			#t2 = item2.tangent(0)
			return nil if p1 == p2
			
			#Direction of rotation (i.e. the normal) is determined by the direction of offset.
			normal = Geom::Vector3d.new(0,0,-offset_amnt).normalize
			
			#Determine the center point. It's one of the two points at the intersections of circles as given below.
			null_vec = Geom::Vector3d.new(1,0,0)
			circ1 = PGL_ArcDef.new(p1,offset_amnt.abs,null_vec,null_vec)
			circ2 = PGL_ArcDef.new(p2,offset_amnt.abs,null_vec,null_vec)
			c_array = PathGeomLib.intersect_circle_circle(circ1,circ2).pts
			return nil if c_array.length == 0
			
			center = nil
			
			#Assume it's the first point, switch to the other if not.
			c1 = c_array[0]
			
			#If 180 degree arc, the only center will be directly between both intersections.
			center = nil
			if c_array.length == 1 then
				center = c1			
			elsif (de_float_error(c1.vector_to(p1).cross(c1.vector_to(p2)).z) > 0) ^ (de_float_error(normal.z) < 0) then
				center = c1
			else
				center = c_array[1]
			end
			
			corner = PGL_ArcDef.new(center,p1.distance(center),center.vector_to(p1),center.vector_to(p2),normal)
			
			#Return the corner.
			begin
				return corner
			rescue
				raise "center: #{center}, p1: #{p1}, c_array: #{c_array.inspect}"
			end
			
		end
		
		def final_pass(failsafe = 0)
			raise "Error: final pass looped indefinitely" if failsafe > 1000
			
			pass_again = false
			@pgl_items.each do |item|
				if item.zero? then
					@pgl_items.delete(item)
					pass_again = true
				end
			end
			
			delete_array = []
			0.upto(@pgl_items.length - 1) do |i|
				j = (i + 1) % @pgl_items.length
				next if j == 0 and not closed? # We don't want to fuse last item to first item unless the path is closed
				
				# Perform any missed fuses
				fuse = @pgl_items[i].fuse!(@pgl_items[j])
				if not fuse then
					puts "------------"
					puts "i: #{@pgl_items[i].describe}"
					puts "j: #{@pgl_items[j].describe}"
					$debug = true
					@pgl_items[i].fuse!(@pgl_items[j])
					$debug = false
					raise "Error: could not fuse during final pass"
				end
				delete_array << @pgl_items[j] if @pgl_items[j].zero?
			end
			
			pass_again = true if delete_array.length > 0
			delete_array.each do |d|
				@pgl_items.delete(d)
			end
			
			final_pass(failsafe + 1) if pass_again
		end
		
		def set_array_precise
			@pgl_items.each do |item|
				item.separation_allowed = 0
			end
		end
		
		def get_subpaths
			return
			intersects = fragment_path
			paths = []
			paths << PGL_Path.new
			@pgl_items.each do |item|
				item = item.clone
				paths.last << item
				
				@pgl_items.each do |compared|
					int = item.intersect(compared)
					if int then
						item2 = item.split(int)
						paths << PGL_Path.new([item2])
					end
				end
			end
			
			return paths
		end
		
		def get_trimmed_paths(original_path,offset_amount)
			
		end
	end

end