module Grays42_GPath

	class PGL_SegmentDef < PGL_Item

		def initialize(start_point,end_point)
			@points = [start_point,end_point]
			@separation_allowed = nil
			raise "Segments must be on a z-flat plane" unless start_point.z == end_point.z
		end
		
		attr_accessor :separation_allowed
		
		def transform!(transformation)
			@points[0] = @points[0].transform(transformation)
			@points[1] = @points[1].transform(transformation)
		end
		
		def describe
			return "Seg (#{start_pt.x},#{start_pt.y})->(#{end_pt.x},#{end_pt.y}), len=#{length}\n"
		end
		
		def draw(entities = Sketchup.active_model.entities)
			return entities.add_line(@points[0].clone,@points[1].clone)
		end
		
		def within_separation_tolerance(reference_index,pt)
			#Edges don't care where they intersect unless they are intersecting other edges, will be handled in the fuse routine.
			return true if @separation_allowed.nil?
			return true if distance_from(pt) <= @separation_allowed
			return false
		end
		
		def ==(seg)
			return false unless seg.class == PGL_SegmentDef
			return true if seg.points == @points
			return false
		end
		
		def offset!(amnt)
			return false if zero?
			through_vec = @points[0].vector_to(@points[1])
			offset_vec = through_vec.cross(Geom::Vector3d.new(0,0,-1))
			offset_vec.length = amnt
			@points.each do |p|
				p.offset!(offset_vec)
			end
		end
		
		def offset(amnt)
			ret = clone
			ret.offset!(amnt)
			return ret
		end
		
		def tangent(index)
			#For segments, it doesn't matter which index we get; the tangent vector will be the same at either point.
			return @points[0].vector_to(@points[1]).normalize
		end
		
		#Is the segment zero-length?
		def zero?
			return true if @points[0] == @points[1]
			return false
		end
		
		def reverse!
			@points.reverse!
		end
		
		def reverse
			ret = self.clone
			ret.reverse!
			return ret
		end
		
		#Splits the two. Edits own values for the first half, returns the second half.
		def split(point)
			raise "Error: cannot split PGL item using a point that is not on the item." unless distance_from(point) == 0
			return nil if point == start_pt or point == end_pt
			second = clone
			second.set_pt(0,point)
			set_pt(1,point)
			return second
		end
		
		def get_halves(point)
			item1 = clone
			item2 = item1.split(point)
			return [item1,item2]
		end
		
		def length
			return @points[0].distance(@points[1])
		end
		
		def intersect(other_item, guaranteed_intersection = false)
			return PathGeomLib.intersect(self, other_item, guaranteed_intersection)
		end
		
		def same_line?(other_segment)
			return false unless other_segment.class == PGL_SegmentDef
			m_same = false
			if other_segment.m_infinity? or m_infinity? then
				m_same = true if m.class == other_segment.m.class and m == other_segment.m
			else
				m_same = true if (other_segment.m - m).to_l == 0
			end
			return true if (other_segment.b - b).to_l == 0 and m_same
			return false
		end
		
		def fuse!(other_item, guaranteed_intersection = false)
			return false if self == other_item
			int = intersect(other_item, guaranteed_intersection)
			return false if int.nil?
			if other_item.class == PGL_SegmentDef and same_line?(other_item) then
				return false unless distance_from(int) == 0 and other_item.distance_from(int) == 0
				#Steal the endpoint and zero out the other line.
				set_pt(1,other_item.end_pt)
				other_item.set_pt(0,other_item.end_pt)
				return true
			elsif other_item.within_separation_tolerance(0,int) then
				set_pt(1,int)
				other_item.set_pt(0,int)
				return true
			end
		end
		
		def clone
			item = PGL_SegmentDef.new(@points[0].clone, @points[1].clone)
			item.separation_allowed = @separation_allowed
			return item
		end
		
		def set_pt(index,pt)
			@points[index] = pt.clone
		end
		
		def start_pt
			return @points[0].clone
		end

		def end_pt
			return @points[1].clone
		end
		
		def points
			return [@points[0].clone,@points[1].clone]
		end
		
		# y = mx + b

		def m
			if @points[0].x == @points[1].x then
				#For some reason Sketchup doesn't like Float::INFINITY, so I'll use this as a placeholder.
				return :infinity
			else
				return ((@points[0].y - @points[1].y)/(@points[0].x - @points[1].x)).to_l ## START HERE, CONVERTING TO LENGTHS
			end
		end
		
		#Separate function ncessary because comparing a Length to a Symbol throws an error
		def m_infinity?
			return true if @points[0].x == @points[1].x
			return false
		end
			
		def b
			return @points[0].x if m_infinity?
			y = @points[0].y
			x = @points[0].x
			b = y - m*x
			return b.to_l
		end
		
		def distance_from(p)
			#Technique directly copied from:
			#http://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
			#Note that the variable names are horrible and I don't fully understand how this works...but it works. Whatever.
			
			v = @points[0]
			w = @points[1]
			
			def dist2(v,w)
				return ((v.x - w.x)**2 + (v.y - w.y)**2).to_l
			end
			
			raise "Can only test point on same plane as segment" unless p.z == v.z
			return dist2(p,v) if zero?
			
			t = (((p.x - v.x) * (w.x - v.x) + (p.y - v.y) * (w.y - v.y)) / (length**2)).to_l
			
			return Math::sqrt(dist2(p,v)).to_l if t < 0
			return Math::sqrt(dist2(p,w)).to_l if t > 1
			return 0.to_l if p.on_line?(v,w)
			return Math::sqrt(p.distance_to_line(v,w)).to_l
		end
		
	end
end