module Grays42_GPath

	#This is a collection of EntityPaths (housing PGL items representing abstract geometry) that have been culled from a Face entity. The purpose is to
	#automatically parse and make these paths available to then be fed into the gcode generator. After being initially fed a face, the FacePaths object does
	#not need that face to continue to exist; all parsing is completed  at the moment of initialization.

	class Solid_Paths
		
		def initialize(group)
			raise "Class error: #{group.class}." unless group.class == Sketchup::Group or group.class == Sketchup::ComponentInstance
			
			@group = group
			
			bottom_faces = get_bottom_faces
			@paths = []
			
			bottom_faces.each do |face|
				@paths = paths + generate_paths_from_face(face)
			end
			
			lock_all_paths
			tsp
		end
		
		
		attr_accessor :paths
		attr_accessor :group
		
		def draw
			@paths.each do |path|
				path.draw
			end
		end
		
		def length
			len = 0
			@paths.each do |path|
				len += path.length
			end
			return len
		end
		
		def lock_all_paths
			@paths.each do |path|
				path.set_array_precise
			end
		end
		
		def offset!(val)
			@paths.each do |path|
				path.offset!(val)
			end
		end
		
		def describe
			ret = "Solid Item \"#{@group.name}\":\n"
			@paths.each do |path|
				ret << path.describe
			end
			return ret
		end
		
		def start_pt
			return @paths[0].start_pt.clone
		end
		
		def end_pt
			return @paths.last.end_pt.clone
		end
		
		def get_bottom_faces
			
			@def = @group if @group.class == Sketchup::Group
			@def = @group.definition if @group.class == Sketchup::ComponentInstance
			
			faces = {}
			
			@def.entities.each do |e|
				faces[e] = 0 if e.class == Sketchup::Face
			end
			
			faces.keys.each do |f|
				
				f.vertices.each do |v|
					faces[f] = faces[f] + v.position.transform(@group.transformation).z
				end
				
				faces[f] = faces[f].to_f / f.vertices.length.to_f unless f.vertices.length == 0
				
				faces.delete(f) unless PathGeomLib::de_float_error(faces[f] - faces.values.min,4) == 0
			end
			
			return faces.keys
			
		end
		
		def generate_path_from_loop(loop)
			path = Entity_Path.new
			
			loop.edges.each do |e|
				path.append_entity(e,@group.transformation)
			end
			
			#explicitly check circles.
			if path.circle? then
				path.reverse! if path.pgl_items[0].direction == :counter_clockwise and loop.outer?
				path.reverse! if path.pgl_items[0].direction == :clockwise and not loop.outer?
			end
			
			path.close!
			
			return path
		end
		
		def generate_paths_from_face(face)
			ret = []
			face.loops.each do |loop|
				ret << generate_path_from_loop(loop)
			end
			return ret
		end
		
		#This implementation of the travelling salesman problem is extremely crude, as the "optimal" solutions are
		#beyond my current understanding without spending an hour or two of research. Basically, it just takes each
		#item's start point and movss to the nearest start point. It dramatically trims down the seek time for pieces
		#that have many interior holes.
		def tsp
			
			#The first path is the "outside" loop.
			new_path_set = [@paths.first]
			@paths.shift
			
			failsafe = 0
			while @paths.length > 0 do
				
				distances = {}
				@paths.each do |path|
					distances[new_path_set.last.start_pt.distance(path.start_pt)] = path
				end
				lowest_dist = distances.keys.min
				path = distances[lowest_dist]
				
				new_path_set << path
				@paths.delete(path)
				failsafe = failsafe + 1
				raise "Error: infinite loop" if failsafe > 10000
			end
			
			#We want the outside loop to be the last thing we cut, so swap the order.
			@paths = new_path_set.reverse
		end

	end
end